unit Unit53;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls;

type
  TForm53 = class(TForm)
    PaintBox1: TPaintBox;
    procedure FormCreate(Sender: TObject);
    procedure PaintBox1Paint(Sender: TObject);
    procedure PaintBox1Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form53: TForm53;

implementation

{$R *.dfm}

const
   SAMPLES_BITS = 10;
   SAMPLES = 1 shl SAMPLES_BITS;
   NB = 512*SAMPLES;
   NBTHREADS = 4;

type
   TData = array [0..SAMPLES-1, 0..NBTHREADS-1] of Integer;
   PData = ^TData;

var
   vcs : TRTLCriticalSection;
   vCounter : Integer;
   vDataCS, vDataMon : TData;

type
   TTestThread = class(TThread)
      Data : PData;
      N : Integer;
      procedure Execute; override;
      procedure LockData; virtual; abstract;
      procedure UnLockData; virtual; abstract;
   end;

   TCSThread = class (TTestThread)
      procedure LockData; override;
      procedure UnLockData; override;
   end;

   TMonitorThread = class (TTestThread)
      procedure LockData; override;
      procedure UnLockData; override;
   end;

// Execute
//
procedure TTestThread.Execute;
begin
   while not Terminated do begin
      LockData;
      try
         if vCounter>=NB then
            Terminate
         else begin
            Inc(vCounter);
            Inc(Data[vCounter shr SAMPLES_BITS, N]);
         end;
      finally
         UnLockData;
      end;
   end;
end;

// LockData
//
procedure TCSThread.LockData;
begin
   EnterCriticalSection(vcs);
end;

// UnLockData
//
procedure TCSThread.UnLockData;
begin
   LeaveCriticalSection(vcs);
end;

// LockData
//
procedure TMonitorThread.LockData;
begin
   System.TMonitor.Enter(Form53);
end;

// UnLockData
//
procedure TMonitorThread.UnLockData;
begin
   System.TMonitor.Exit(Form53);
end;

procedure TForm53.FormCreate(Sender: TObject);
var
   qs, qe, qf : Int64;
   i : Integer;
   threads : array [0..NBTHREADS-1] of TTestThread;
begin
   SetProcessAffinityMask(GetCurrentProcess, Cardinal(-1));

   InitializeCriticalSection(vcs);

   QueryPerformanceFrequency(qf);

   FillChar(vDataCS, SizeOf(vDataCS), 0);

   QueryPerformanceCounter(qs);

   vCounter:=0;
   for i:=0 to High(threads) do begin
      threads[i]:=TCSThread.Create;
      threads[i].N:=i;
      threads[i].Data:=@vDataCS;
   end;
   for i:=0 to High(threads) do begin
      threads[i].WaitFor;
      threads[i].Free;
   end;

   QueryPerformanceCounter(qe);

   Caption:=Format('%.3f s', [(qe-qs)/qf]);

   FillChar(vDataMon, SizeOf(vDataMon), 0);

   QueryPerformanceCounter(qs);

   vCounter:=0;
   for i:=0 to High(threads) do begin
      threads[i]:=TMonitorThread.Create;
      threads[i].N:=i;
      threads[i].Data:=@vDataMon;
   end;
   for i:=0 to High(threads) do begin
      threads[i].WaitFor;
      threads[i].Free;
   end;

   QueryPerformanceCounter(qe);

   PaintBox1.Invalidate;

   Caption:=Caption+Format(' %.3f s', [(qe-qs)/qf]);

   DeleteCriticalSection(vcs);
end;

procedure TForm53.PaintBox1Click(Sender: TObject);
begin
   FormCreate(Self);
   PaintBox1.Invalidate;
end;

procedure TForm53.PaintBox1Paint(Sender: TObject);
var
   i, j, k : Integer;
   h : Integer;
   r : TRect;
   canvas : TCanvas;
begin
   canvas:=PaintBox1.Canvas;
   canvas.Brush.Style:=bsSolid;
   canvas.Brush.Color:=clWhite;
   canvas.FillRect(PaintBox1.ClientRect);

   h:=32 div NBTHREADS;

   for i:=0 to 511 do begin
      r.Left:=2*i;
      r.Right:=2*i+2;
      for j:=0 to NBTHREADS do begin
         r.Top:=j*h;
         r.Bottom:=j*h+h;
         k:=vDataCS[i, j]*255*NBTHREADS div SAMPLES div 2;
         canvas.Brush.Color:=RGB(k, k, k);
         canvas.FillRect(r);

         r.Top:=j*h+NBTHREADS*h+16;
         r.Bottom:=j*h+NBTHREADS*h+16+h;
         k:=vDataMon[i, j]*255*NBTHREADS div SAMPLES div 2;
         canvas.Brush.Color:=RGB(k, k, k);
         canvas.FillRect(r);
      end;
   end;
end;

end.
